'use strict';

var react = require('react');

const defaultReturn = {
    isValid: false,
    value: null,
};
function getRadioValue(options) {
    return Array.isArray(options)
        ? options.reduce((previous, { ref: { name, checked, value } }) => checked
            ? {
                isValid: true,
                value,
            }
            : previous, defaultReturn)
        : defaultReturn;
}

function getMultipleSelectValue(options) {
    return options.filter(({ selected }) => selected).map(({ value }) => value);
}

function isRadioInput(type) {
    return type === 'radio';
}

function getFieldValue(fields, { type, name, options, checked, value }) {
    if (isRadioInput(type)) {
        return getRadioValue(fields[name].options).value;
    }
    else if (type === 'select-multiple') {
        return getMultipleSelectValue(options);
    }
    else if (type === 'checkbox') {
        return checked;
    }
    return value;
}

function getFieldsValues(fields, filedNames) {
    return Object.values(fields).reduce((previous, data) => {
        if (Array.isArray(data))
            return;
        const { ref, ref: { name } } = data;
        const value = getFieldValue(fields, ref);
        if (typeof filedNames === 'string') {
            if (name === filedNames) {
                return value;
            }
            return previous;
        }
        const copy = Object.assign({}, (previous || {}));
        if (Array.isArray(filedNames)) {
            if (filedNames.includes(name)) {
                copy[name] = value;
            }
        }
        else {
            copy[name] = value;
        }
        return copy;
    }, undefined);
}

const DATE_INPUTS = ['date', 'time', 'month', 'datetime', 'datetime-local', 'week'];
const STRING_INPUTS = ['text', 'email', 'password', 'search', 'tel', 'url', 'textarea'];

function getValueAndMessage(item) {
    return {
        value: typeof item === 'object' && item.value ? item.value : item,
        message: typeof item === 'object' && item.message ? item.message : true,
    };
}

var validateField = async ({ ref, ref: { type, value, name, checked }, options, required, maxLength, minLength, min, max, pattern, validate, }, fields) => {
    const copy = {};
    const isRadio = isRadioInput(type);
    if (required &&
        ((type === 'checkbox' && !checked) ||
            (!isRadio && type !== 'checkbox' && value === '') ||
            (isRadio && !getRadioValue(fields[name].options).isValid))) {
        copy[name] = {
            type: 'required',
            message: required,
            // @ts-ignore
            ref: isRadio ? fields[name].options[0].ref : ref,
        };
        return copy;
    }
    // min and max section
    if (min || max) {
        let exceedMax;
        let exceedMin;
        const valueNumber = parseFloat(value);
        const { value: maxValue, message: maxMessage } = getValueAndMessage(max);
        const { value: minValue, message: minMessage } = getValueAndMessage(min);
        if (type === 'number') {
            exceedMax = maxValue && valueNumber > maxValue;
            exceedMin = minValue && valueNumber < minValue;
        }
        else if (DATE_INPUTS.includes(type)) {
            if (typeof maxValue === 'string')
                exceedMax = maxValue && new Date(value) > new Date(maxValue);
            if (typeof minValue === 'string')
                exceedMin = minValue && new Date(value) < new Date(minValue);
        }
        if (exceedMax) {
            copy[name] = Object.assign({}, copy[name], { type: 'max', message: maxMessage, ref });
            return copy;
        }
        if (exceedMin) {
            copy[name] = Object.assign({}, copy[name], { type: 'min', message: minMessage, ref });
            return copy;
        }
    }
    if ((maxLength || minLength) && STRING_INPUTS.includes(type)) {
        const { value: maxLengthValue, message: maxLengthMessage } = getValueAndMessage(maxLength);
        const { value: minLengthValue, message: minLengthMessage } = getValueAndMessage(minLength);
        const exceedMax = maxLength && value.toString().length > maxLengthValue;
        const exceedMin = minLength && value.toString().length < minLengthValue;
        if (exceedMax) {
            copy[name] = Object.assign({}, copy[name], { type: 'maxLength', message: maxLengthMessage, ref });
            return copy;
        }
        if (exceedMin) {
            copy[name] = Object.assign({}, copy[name], { type: 'minLength', message: minLengthMessage, ref });
            return copy;
        }
    }
    if (pattern) {
        const { value: patternValue, message: patternMessage } = getValueAndMessage(pattern);
        if (patternValue instanceof RegExp && !patternValue.test(value)) {
            copy[name] = Object.assign({}, copy[name], { type: 'pattern', message: patternMessage, ref });
            return copy;
        }
    }
    if (validate) {
        const fieldValue = isRadio ? getRadioValue(options).value : value;
        if (typeof validate === 'function') {
            const result = await validate(fieldValue);
            if (typeof result !== 'boolean' || !result) {
                copy[name] = Object.assign({}, copy[name], { type: 'validate', message: result || true, ref: isRadio && options ? options[0].ref : ref });
                return copy;
            }
        }
        else if (typeof validate === 'object') {
            const result = await new Promise(resolve => {
                const values = Object.entries(validate);
                values.reduce(async (previous, [key, validate], index) => {
                    const result = typeof validate === 'function' && (await validate(fieldValue));
                    const lastChild = values.length - 1 === index;
                    if (Object.keys(previous).length)
                        return lastChild ? resolve(previous) : previous;
                    if (typeof result !== 'boolean' || !result) {
                        const temp = {
                            type: key,
                            message: result || true,
                        };
                        return lastChild ? resolve(temp) : temp;
                    }
                    if (lastChild)
                        return resolve(previous);
                }, {});
            });
            if (Object.keys(result).length) {
                copy[name] = Object.assign({}, copy[name], { ref: isRadio && options ? options[0].ref : ref }, result);
                return copy;
            }
        }
    }
    return copy;
};

var removeAllEventListeners = (ref, validateWithStateUpdate) => {
    ref.removeEventListener('input', validateWithStateUpdate);
    ref.removeEventListener('change', validateWithStateUpdate);
    ref.removeEventListener('blur', validateWithStateUpdate);
};

function findMissDomAndClean({ target: { ref, ref: { name, type }, mutationWatcher, options, }, fields, validateWithStateUpdate, forceDelete = false, }) {
    if (isRadioInput(type) && options) {
        options.forEach(({ ref }, index) => {
            if (!document.body.contains(ref) && fields[name] && fields[name].options && fields[name].options[index]) {
                removeAllEventListeners(fields[name].options[index], validateWithStateUpdate);
                fields[name].options[index].mutationWatcher.disconnect();
                fields[name].options.splice(index, 1);
            }
        });
        if (!fields[name].options.length) {
            delete fields[name];
            return fields;
        }
    }
    else if (ref && (!document.body.contains(ref) || forceDelete)) {
        removeAllEventListeners(ref, validateWithStateUpdate);
        if (mutationWatcher)
            mutationWatcher.disconnect();
        delete fields[name];
        return fields;
    }
    return fields;
}

function onDomRemove(element, onDetachCallback) {
    const observer = new MutationObserver(() => {
        function isDetached(el) {
            if (el.parentNode === window.document) {
                return false;
            }
            else if (el.parentNode === null) {
                return true;
            }
            return isDetached(el.parentNode);
        }
        if (isDetached(element)) {
            observer.disconnect();
            onDetachCallback();
        }
    });
    observer.observe(window.document, {
        childList: true,
        subtree: true,
    });
    return observer;
}

function attachEventListeners({ mode, fields, watchFields, radioOptionIndex, ref, type, name, validateWithStateUpdate, }) {
    const field = fields[name];
    const isOnChange = mode === 'onChange' || watchFields[ref.name];
    const isOnBlur = mode === 'onBlur';
    if (!field || (!isOnChange && !isOnBlur))
        return;
    const isRadio = isRadioInput(type);
    const event = isOnChange ? (isRadio ? 'change' : 'input') : 'blur';
    if (isRadio) {
        const options = field.options;
        const attachedEvents = options[radioOptionIndex].eventAttached;
        if (!options[radioOptionIndex])
            return;
        if (!attachedEvents || (attachedEvents && !attachedEvents.includes(event))) {
            options[radioOptionIndex].ref.addEventListener(event, validateWithStateUpdate);
            options[radioOptionIndex].eventAttached = [...attachedEvents || [], event];
        }
    }
    else {
        if (!field.eventAttached || (field.eventAttached && !field.eventAttached.includes(event))) {
            ref.addEventListener(event, validateWithStateUpdate);
            field.eventAttached = [...field.eventAttached || [], event];
        }
    }
}

function parseErrorSchema(error) {
    return error.inner.reduce((previous, current, index) => {
        if (!previous[current.path])
            previous[current.path] = {};
        previous[current.path][current.type] = error.errors[index];
        return previous;
    }, {});
}
async function validateWithSchema(ValidationSchema, data) {
    try {
        await ValidationSchema.validate(data, { abortEarly: false });
        return undefined;
    }
    catch (e) {
        return parseErrorSchema(e);
    }
}

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

function __rest(s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) if (e.indexOf(p[i]) < 0)
            t[p[i]] = s[p[i]];
    return t;
}

var omitRefs = (errors) => Object.entries(errors).reduce((previous, [key, values]) => {
    if (!values.ref) {
        previous[key] = values;
        return previous;
    }
    const rest = __rest(values, ["ref"]);
    previous[key] = rest;
    return previous;
}, {});

function combineFieldValues(data) {
    const output = Object.entries(data).reduce((previous, [key, value]) => {
        const arrayIndex = key.match(/\[\d+\]$/gi);
        if (arrayIndex) {
            const index = arrayIndex[0].slice(1, -1);
            const filedName = key.substr(0, key.indexOf('['));
            if (!previous[filedName])
                previous[filedName] = [];
            previous[filedName][index] = value;
        }
        else {
            previous[key] = value;
        }
        return previous;
    }, {});
    return Object.entries(output).reduce((previous, [key, value]) => {
        if (Array.isArray(value)) {
            previous[key] = value.filter(Boolean);
            return previous;
        }
        previous[key] = value;
        return previous;
    }, {});
}

function useForm({ mode, validationSchema } = {
    mode: 'onSubmit',
}) {
    const fieldsRef = react.useRef({});
    const errorMessagesRef = react.useRef({});
    const watchFieldsRef = react.useRef({});
    const [errors, setErrors] = react.useState({});
    async function validateWithStateUpdate({ target: { name }, type }) {
        const ref = fieldsRef.current[name];
        const error = await validateField(ref, fieldsRef.current);
        const errorMessage = errorMessagesRef.current;
        if (errorMessage[name] !== error[name] ||
            mode === 'onChange' ||
            (mode === 'onBlur' && type === 'blur') ||
            watchFieldsRef.current[name]) {
            const copy = Object.assign({}, errorMessage, error);
            if (!error[name])
                delete copy[name];
            errorMessagesRef.current = copy;
            setErrors(copy);
        }
    }
    function removeReferenceAndEventListeners(data, forceDelete = false) {
        findMissDomAndClean({
            target: data,
            fields: fieldsRef.current,
            validateWithStateUpdate,
            forceDelete,
        });
    }
    function registerIntoAllFields(elementRef, data = {}) {
        if (elementRef && !elementRef.name) {
            console.warn('Oops missing the name for field:', elementRef);
            return;
        }
        let radioOptionIndex;
        const inputData = Object.assign({}, data, { ref: elementRef });
        const { ref, required, validate, ref: { name, type, value }, } = inputData;
        const fields = fieldsRef.current;
        if (isRadioInput(type)) {
            if (!fields[name]) {
                fields[name] = { options: [], required, validate, ref: { type: 'radio', name } };
            }
            if (!fields[name].validate && validate) {
                fields[name].validate = validate;
            }
            const options = fields[name].options || [];
            radioOptionIndex = options.findIndex(({ ref }) => value === ref.value);
            if (radioOptionIndex > -1) {
                options[radioOptionIndex] = Object.assign({}, options[radioOptionIndex], inputData);
            }
            else {
                options.push(Object.assign({}, inputData, { mutationWatcher: onDomRemove(ref, () => removeReferenceAndEventListeners(inputData, true)) }));
                radioOptionIndex = options.length - 1;
            }
        }
        else {
            const isInitialCreate = !fields[name];
            fields[name] = Object.assign({}, fields[name], inputData);
            if (isInitialCreate) {
                fields[name].mutationWatcher = onDomRemove(ref, () => removeReferenceAndEventListeners(inputData, true));
            }
        }
        attachEventListeners({
            fields,
            watchFields: watchFieldsRef,
            ref,
            type,
            radioOptionIndex,
            name,
            mode,
            validateWithStateUpdate,
        });
    }
    function register(data) {
        if (!data)
            return;
        if (data instanceof HTMLElement) {
            registerIntoAllFields(data);
        }
        return ref => {
            if (ref) {
                registerIntoAllFields(ref, data);
            }
        };
    }
    function watch(filedNames, defaultValue) {
        const watchFields = watchFieldsRef.current;
        if (typeof filedNames === 'string') {
            if (!watchFields[filedNames])
                watchFields[filedNames] = true;
        }
        else if (Array.isArray(filedNames)) {
            filedNames.forEach(name => {
                if (!watchFields[name])
                    return;
                watchFields[name] = true;
            });
        }
        else {
            Object.values(fieldsRef.current).forEach(({ ref }) => {
                if (!ref || !watchFields[ref.name])
                    return;
                watchFields[ref.name] = true;
            });
        }
        const result = getFieldsValues(fieldsRef.current, filedNames);
        return result === undefined ? defaultValue : result;
    }
    const handleSubmit = (callback) => async (e) => {
        e.preventDefault();
        e.persist();
        let fieldErrors;
        let fieldValues;
        const fields = fieldsRef.current;
        const currentFieldValues = Object.values(fields);
        const fieldsLength = currentFieldValues.length;
        if (validationSchema) {
            fieldValues = currentFieldValues.reduce((previous, { ref }) => {
                previous[ref.name] = getFieldValue(fields, ref);
                return previous;
            }, {});
            fieldErrors = await validateWithSchema(validationSchema, fieldValues);
            if (fieldErrors === undefined) {
                callback(combineFieldValues(fieldValues), e);
                return;
            }
        }
        else {
            const result = await new Promise(resolve => currentFieldValues.reduce(async (previous, field, index) => {
                const resolvedPrevious = await previous;
                const { ref, ref: { name, type }, options, } = field;
                const lastChild = fieldsLength - 1 === index;
                removeReferenceAndEventListeners(field);
                if (!fields[name])
                    return lastChild ? resolve(resolvedPrevious) : resolvedPrevious;
                const fieldError = await validateField(field, fields);
                const hasError = fieldError && fieldError[name];
                if (!hasError) {
                    resolvedPrevious.values[name] = getFieldValue(fields, ref);
                    return lastChild ? resolve(resolvedPrevious) : resolvedPrevious;
                }
                if (isRadioInput(type) && Array.isArray(options)) {
                    options.forEach(option => {
                        if (option.eventAttached && option.eventAttached.includes('change'))
                            return;
                        option.ref.addEventListener('change', validateWithStateUpdate);
                        option.eventAttached = [...(option.eventAttached || []), 'change'];
                    });
                    // @ts-ignore
                }
                else if (!field.eventAttached || !field.eventAttached.includes('input')) {
                    ref.addEventListener('input', validateWithStateUpdate);
                    field.eventAttached = [...(field.eventAttached || []), 'input'];
                }
                resolvedPrevious.errors = Object.assign({}, (resolvedPrevious.localErrors || []), fieldError);
                return lastChild ? resolve(resolvedPrevious) : resolvedPrevious;
            }, Promise.resolve({
                errors: {},
                values: {},
            })));
            fieldErrors = result.errors;
            fieldValues = result.values;
        }
        if (JSON.stringify(omitRefs(errorMessagesRef.current)) !== JSON.stringify(omitRefs(fieldErrors))) {
            setErrors(fieldErrors);
            errorMessagesRef.current = fieldErrors;
        }
        if (!Object.values(fieldErrors).length)
            callback(combineFieldValues(fieldValues), e);
    };
    react.useEffect(() => () => {
        fieldsRef.current &&
            Object.values(fieldsRef.current).forEach(({ ref, options }) => isRadioInput(ref.type) && Array.isArray(options)
                ? options.forEach(({ ref }) => removeAllEventListeners(ref, validateWithStateUpdate))
                : removeAllEventListeners(ref, validateWithStateUpdate));
        fieldsRef.current = {};
        watchFieldsRef.current = {};
        errorMessagesRef.current = {};
        setErrors({});
    }, [mode]);
    return {
        register,
        handleSubmit,
        errors,
        watch,
    };
}

module.exports = useForm;
